
em_coherence_we = r"""
@inproceedings{DBLP:conf/ldk/BelfordG19,
  author    = {Mark Belford and
               Derek Greene},
  title     = {Comparison of Embedding Techniques for Topic Modeling Coherence Measures},
  booktitle = {Proceedings of the Poster Session of the 2nd Conference on Language,
               Data and Knowledge {(LDK} 2019), Leipzig, Germany, May 21, 2019},
  pages     = {1--5},
  year      = {2019},
  crossref  = {DBLP:conf/ldk/2019p},
  url       = {http://ceur-ws.org/Vol-2402/paper1.pdf},
  timestamp = {Wed, 12 Feb 2020 16:44:16 +0100},
  biburl    = {https://dblp.org/rec/conf/ldk/BelfordG19.bib},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}
"""

em_word_embeddings_pc = r"""
@inproceedings{DBLP:conf/emnlp/DingNX18,
  author    = {Ran Ding and
               Ramesh Nallapati and
               Bing Xiang},
  editor    = {Ellen Riloff and
               David Chiang and
               Julia Hockenmaier and
               Jun'ichi Tsujii},
  title     = {Coherence-Aware Neural Topic Modeling},
  booktitle = {Proceedings of the 2018 Conference on Empirical Methods in Natural
               Language Processing, Brussels, Belgium, October 31 - November 4, 2018},
  pages     = {830--836},
  publisher = {Association for Computational Linguistics},
  year      = {2018},
  url       = {https://doi.org/10.18653/v1/d18-1096},
  doi       = {10.18653/v1/d18-1096},
  timestamp = {Tue, 28 Jan 2020 10:28:21 +0100},
  biburl    = {https://dblp.org/rec/conf/emnlp/DingNX18.bib},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}
"""

em_coherence = r"""
@inproceedings{DBLP:conf/wsdm/RoderBH15,
  author    = {Michael R{\"{o}}der and
               Andreas Both and
               Alexander Hinneburg},
  editor    = {Xueqi Cheng and
               Hang Li and
               Evgeniy Gabrilovich and
               Jie Tang},
  title     = {Exploring the Space of Topic Coherence Measures},
  booktitle = {Proceedings of the Eighth {ACM} International Conference on Web Search
               and Data Mining, {WSDM} 2015, Shanghai, China, February 2-6, 2015},
  pages     = {399--408},
  publisher = {{ACM}},
  year      = {2015},
  url       = {https://doi.org/10.1145/2684822.2685324},
  doi       = {10.1145/2684822.2685324},
  timestamp = {Tue, 21 May 2019 11:38:33 +0200},
  biburl    = {https://dblp.org/rec/conf/wsdm/RoderBH15.bib},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}
"""

em_topic_diversity = r"""
@article{DBLP:journals/corr/abs-1907-04907,
  author    = {Adji B. Dieng and
               Francisco J. R. Ruiz and
               David M. Blei},
  title     = {Topic Modeling in Embedding Spaces},
  journal   = {CoRR},
  volume    = {abs/1907.04907},
  year      = {2019},
  url       = {http://arxiv.org/abs/1907.04907},
  archivePrefix = {arXiv},
  eprint    = {1907.04907},
  timestamp = {Wed, 17 Jul 2019 10:27:36 +0200},
  biburl    = {https://dblp.org/rec/journals/corr/abs-1907-04907.bib},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}
"""

em_topic_significance = r"""
@inproceedings{DBLP:conf/pkdd/AlSumaitBGD09,
  author    = {Loulwah AlSumait and
               Daniel Barbar{\'{a}} and
               James Gentle and
               Carlotta Domeniconi},
  editor    = {Wray L. Buntine and
               Marko Grobelnik and
               Dunja Mladenic and
               John Shawe{-}Taylor},
  title     = {Topic Significance Ranking of {LDA} Generative Models},
  booktitle = {Machine Learning and Knowledge Discovery in Databases, European Conference,
               {ECML} {PKDD} 2009, Bled, Slovenia, September 7-11, 2009, Proceedings,
               Part {I}},
  series    = {Lecture Notes in Computer Science},
  volume    = {5781},
  pages     = {67--82},
  publisher = {Springer},
  year      = {2009},
  url       = {https://doi.org/10.1007/978-3-642-04180-8\_22},
  doi       = {10.1007/978-3-642-04180-8\_22},
  timestamp = {Tue, 14 May 2019 10:00:47 +0200},
  biburl    = {https://dblp.org/rec/conf/pkdd/AlSumaitBGD09.bib},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}
"""

models_HDP = r"""
@inproceedings{DBLP:conf/nips/TehJBB04,
  author    = {Yee Whye Teh and
               Michael I. Jordan and
               Matthew J. Beal and
               David M. Blei},
  title     = {Sharing Clusters among Related Groups: Hierarchical Dirichlet Processes},
  booktitle = {Advances in Neural Information Processing Systems 17 [Neural Information
               Processing Systems, {NIPS} 2004, December 13-18, 2004, Vancouver,
               British Columbia, Canada]},
  pages     = {1385--1392},
  year      = {2004},
  url       = {http://papers.nips.cc/paper/2698-sharing-clusters-among-related-groups-hierarchical-dirichlet-processes},
  timestamp = {Fri, 06 Mar 2020 16:59:17 +0100},
  biburl    = {https://dblp.org/rec/conf/nips/TehJBB04.bib},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}
            """

models_LDA = r"""
@inproceedings{DBLP:conf/nips/BleiNJ01,
  author    = {David M. Blei and
               Andrew Y. Ng and
               Michael I. Jordan},
  editor    = {Thomas G. Dietterich and
               Suzanna Becker and
               Zoubin Ghahramani},
  title     = {Latent Dirichlet Allocation},
  booktitle = {Advances in Neural Information Processing Systems 14 [Neural Information
               Processing Systems: Natural and Synthetic, {NIPS} 2001, December 3-8,
               2001, Vancouver, British Columbia, Canada]},
  pages     = {601--608},
  publisher = {{MIT} Press},
  year      = {2001},
  url       = {http://papers.nips.cc/paper/2070-latent-dirichlet-allocation},
  timestamp = {Thu, 12 Mar 2020 11:31:34 +0100},
  biburl    = {https://dblp.org/rec/conf/nips/BleiNJ01.bib},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}
            """

models_NMF = r"""
@article{DBLP:journals/tsp/ZhaoT17,
  author    = {Renbo Zhao and
               Vincent Y. F. Tan},
  title     = {Online Nonnegative Matrix Factorization With Outliers},
  journal   = {{IEEE} Trans. Signal Process.},
  volume    = {65},
  number    = {3},
  pages     = {555--570},
  year      = {2017},
  url       = {https://doi.org/10.1109/TSP.2016.2620967},
  doi       = {10.1109/TSP.2016.2620967},
  timestamp = {Tue, 10 Mar 2020 10:51:48 +0100},
  biburl    = {https://dblp.org/rec/journals/tsp/ZhaoT17.bib},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}
            """

sources_dblp_M10 = r"""@inproceedings{DBLP:conf/ijcai/PanWZZW16,
  author    = {Shirui Pan and
               Jia Wu and
               Xingquan Zhu and
               Chengqi Zhang and
               Yang Wang},
  editor    = {Subbarao Kambhampati},
  title     = {Tri-Party Deep Network Representation},
  booktitle = {Proceedings of the Twenty-Fifth International Joint Conference on
               Artificial Intelligence, {IJCAI} 2016, New York, NY, USA, 9-15 July
               2016},
  pages     = {1895--1901},
  publisher = {{IJCAI/AAAI} Press},
  year      = {2016},
  url       = {http://www.ijcai.org/Abstract/16/271},
  timestamp = {Tue, 20 Aug 2019 16:19:21 +0200},
  biburl    = {https://dblp.org/rec/conf/ijcai/PanWZZW16.bib},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}"""


em_f1_score= r"""@
}"""
